from emepy.eme import EME
from emepy.geometries import Geometry, DynamicPolygon
from emepy.source import Source
from emepy.monitors import Monitor
import numpy as np
from matplotlib import pyplot as plt


class Optimization(object):
    """Optimizatoin objects store geometries and can manipulate design regions. Essentially, they form the methods needed for running shape optimizations"""

    def __init__(self, eme: "EME", geometries: list = [], mesh_z: int = 100) -> None:
        """Creates an instance of Optimization for running shape optimization"""
        self.eme = eme
        self.geometries = geometries
        self.mesh_z = mesh_z
        self.start()

    def add_geometry(self, geometry: "Geometry") -> None:
        """Adds a Geometry object to the optimization"""
        self.geometries.append(geometry)

    def add_geometries(self, *geometries) -> None:
        """Adds any number of Geometry objects as arguments to the optimization"""
        for geometry in geometries:
            self.geometries.append(geometry)

    def get_design(self) -> list:
        """Returns the design region associated with all design geometries in the optimization"""
        design = []
        for geometry in self.geometries:
            if isinstance(geometry, DynamicPolygon):
                d = (
                    geometry.get_design()
                    if not isinstance(geometry.get_design(), np.ndarray)
                    else geometry.get_design().tolist()
                )
                design += d

        return design

    def get_design_readable(self, dimensions: int = 2) -> list:
        """Returns the design region associated with all design geometries in the optimization in a readable form (a tuple in the form (x,z) or (x,y,z)"""
        design = self.get_design()
        if not len(design):
            return [], [] if dimensions == 2 else [], [], []
        elif dimensions == 2:
            x, z = (design[::2], design[1::2])
            return x, z
        elif dimensions == 3:
            x, y, z = (design[::3], design[1::3], design[2::3])
            return x, y, z
        else:
            return []

    def set_design(self, design: list) -> None:
        """Sets the design region provided for the entire system of design regions"""
        remaining_design = design[:]
        for geometry in self.geometries:
            if isinstance(geometry, DynamicPolygon):
                length = len(geometry)
                if length:
                    geometry.set_design(remaining_design[:length])
                    remaining_design = remaining_design[length:] if length < len(remaining_design) else []

    def set_design_readable(
        self, design_x: list = [], design_y: list = [], design_z: list = [], dimensions: int = 2
    ) -> None:
        """Sets the design region provided for the entire system of design regions using readable coordinates"""
        design = []
        if dimensions == 2:
            for x, z in zip(design_x, design_z):
                design.append(x)
                design.append(z)
        elif dimensions == 3:
            for x, y, z in zip(design_x, design_y, design_z):
                design.append(x)
                design.append(y)
                design.append(z)
        if len(design):
            self.set_design(design)

    def start(self) -> None:
        """Initializes the EME"""
        layers = [layer for geometry in self.geometries for layer in geometry]
        self.eme.reset(parallel=self.eme.parallel, configure_parallel=False)
        self.eme.add_layers(*layers)

    def update_eme(self) -> None:
        """Updades the eme object with geometric changes"""
        self.start()

    def get_n(self, grid_x: "np.ndarray", grid_y: "np.ndarray", grid_z: "np.ndarray") -> "np.ndarray":
        """Currently returns the n for the first design region"""
        for geometry in self.geometries:
            if isinstance(geometry, DynamicPolygon):
                return geometry.get_n(grid_x, grid_y, grid_z - grid_z[0])

    def gradient(self, grid_x: "np.ndarray", grid_y, grid_z: "np.ndarray", dp=1e-10) -> "np.ndarray":

        """Computes the gradient A_u using a finite difference"""

        # Get initial design
        design = self.get_design()

        # Final jacobian setup
        jacobian = np.zeros(
            (3, 3, grid_x.shape[0] - 1, grid_y.shape[0] - 1, grid_z.shape[0] - 1, len(design)), dtype=complex
        )

        # Get gradients
        for i, d in enumerate(design):

            # Step up
            design[i] = d + dp
            self.set_design(design)

            # Compute up A
            A_up = self.get_n(grid_x, grid_y, grid_z) ** 2

            # Step down
            design[i] = d - 2 * dp
            self.set_design(design)

            # Compute down A
            A_down = self.get_n(grid_x, grid_y, grid_z) ** 2

            # Compute gradient
            gradient = (A_up - A_down) / (2 * dp)

            # Revert step
            design[i] = d + dp
            self.set_design(design)

            # Assign gradient
            jacobian[0, 0, :, :, :, i] = gradient
            jacobian[1, 1, :, :, :, i] = gradient
            jacobian[2, 2, :, :, :, i] = gradient

        return jacobian

    def get_design_region(self) -> tuple:
        z_start, z_end = (0, 0)
        for geometry in self.geometries:
            if isinstance(geometry, DynamicPolygon):
                z_end += geometry.length
                break
            else:
                z_start += geometry.length
                z_end += geometry.length
        return z_start, z_end

    def forward_run(self) -> tuple:
        """Computes the forward run for the adjoint formulation"""

        # Clear the eme and ensure design is inside
        self.start()

        # Find where monitor should be in range of only the design region
        z_start, z_end = self.get_design_region()

        # Create forward source and monitor
        source = Source(z=0.2500001, mode_coeffs=[1], k=1)  # Hard coded
        forward_monitor = self.eme.add_monitor(axes="xyz", mesh_z=self.mesh_z, sources=[source])

        # FOM monitor
        source = Source(z=0.25, mode_coeffs=[1], k=1)  # Hard coded
        fom_monitor = self.eme.add_monitor(axes="xy", location=2.75, sources=[source])

        # Adjoint source and monitor
        source = Source(z=2.75, mode_coeffs=[1], k=-1)
        adjoint_monitor = self.eme.add_monitor(axes="xyz", mesh_z=self.mesh_z, sources=[source])

        # Run eme
        self.eme.propagate()

        # Get near results
        grid_x, grid_y, grid_z, field_x = forward_monitor.get_array("Ex", z_range=(z_start, z_end))
        field_x = 0.125 * (
            field_x[1:, 1:, 1:]
            + field_x[1:, :-1, 1:]
            + field_x[:-1, 1:, 1:]
            + field_x[:-1, :-1, 1:]
            + field_x[1:, 1:, :-1]
            + field_x[1:, :-1, :-1]
            + field_x[:-1, 1:, :-1]
            + field_x[:-1, :-1, :-1]
        )
        field_y = forward_monitor.get_array("Ey", z_range=(z_start, z_end))[3]
        field_y = 0.125 * (
            field_y[1:, 1:, 1:]
            + field_y[1:, :-1, 1:]
            + field_y[:-1, 1:, 1:]
            + field_y[:-1, :-1, 1:]
            + field_y[1:, 1:, :-1]
            + field_y[1:, :-1, :-1]
            + field_y[:-1, 1:, :-1]
            + field_y[:-1, :-1, :-1]
        )
        field_z = forward_monitor.get_array("Ez", z_range=(z_start, z_end))[3]
        field_z = 0.125 * (
            field_z[1:, 1:, 1:]
            + field_z[1:, :-1, 1:]
            + field_z[:-1, 1:, 1:]
            + field_z[:-1, :-1, 1:]
            + field_z[1:, 1:, :-1]
            + field_z[1:, :-1, :-1]
            + field_z[:-1, 1:, :-1]
            + field_z[:-1, :-1, :-1]
        )
        field = np.array([field_x, field_y, field_z])

        forward_results = (grid_x, grid_y, grid_z, field, forward_monitor, fom_monitor)

        # Save adjoint results
        a_grid_x, a_grid_y, a_grid_z, a_field_x = adjoint_monitor.get_array("Ex", z_range=(z_start, z_end))
        a_field_x = 0.125 * (
            a_field_x[1:, 1:, 1:]
            + a_field_x[1:, :-1, 1:]
            + a_field_x[:-1, 1:, 1:]
            + a_field_x[:-1, :-1, 1:]
            + a_field_x[1:, 1:, :-1]
            + a_field_x[1:, :-1, :-1]
            + a_field_x[:-1, 1:, :-1]
            + a_field_x[:-1, :-1, :-1]
        )
        a_field_y = adjoint_monitor.get_array("Ey", z_range=(z_start, z_end))[3]
        a_field_y = 0.125 * (
            a_field_y[1:, 1:, 1:]
            + a_field_y[1:, :-1, 1:]
            + a_field_y[:-1, 1:, 1:]
            + a_field_y[:-1, :-1, 1:]
            + a_field_y[1:, 1:, :-1]
            + a_field_y[1:, :-1, :-1]
            + a_field_y[:-1, 1:, :-1]
            + a_field_y[:-1, :-1, :-1]
        )
        a_field_z = adjoint_monitor.get_array("Ez", z_range=(z_start, z_end))[3]
        a_field_z = 0.125 * (
            a_field_z[1:, 1:, 1:]
            + a_field_z[1:, :-1, 1:]
            + a_field_z[:-1, 1:, 1:]
            + a_field_z[:-1, :-1, 1:]
            + a_field_z[1:, 1:, :-1]
            + a_field_z[1:, :-1, :-1]
            + a_field_z[:-1, 1:, :-1]
            + a_field_z[:-1, :-1, :-1]
        )
        a_field = np.array([a_field_x, a_field_y, a_field_z])

        self.adjoint_results = (a_grid_x, a_grid_y, a_grid_z, a_field, adjoint_monitor)

        return forward_results

    def overlap(self, E1x, E1y, H1x, H1y, E2x, E2y, H2x, H2y, x, y):
        term1 = np.conj(E1x) * H2y - np.conj(E1y) * H2x
        term2 = E2x * np.conj(H1y) - E2y * np.conj(H1x)
        return np.trapz(np.trapz((term1 + term2), x), y)

    def objective_gradient(self, monitor: "Monitor"):
        """Computes the objective function gradient to the sources for the adjoint formulation"""

        # Compute power in end
        x, y, Ex = monitor.get_array(component="Ex")
        _, _, Ey = monitor.get_array(component="Ey")
        _, _, Hx = monitor.get_array(component="Hx")
        _, _, Hy = monitor.get_array(component="Hy")

        # Reference mode
        if self.eme.am_master():
            reference_mode = self.eme.activated_layers[0][-1].modes[0]
            r_Ex = reference_mode.Ex
            r_Ey = reference_mode.Ey
            r_Hx = reference_mode.Hx
            r_Hy = reference_mode.Hy
            norm = np.sqrt(np.abs(self.overlap(r_Ex, r_Ey, r_Hx, r_Hy, r_Ex, r_Ey, r_Hx, r_Hy, x, y)))
            r_Ex = r_Ex / norm
            r_Ey = r_Ey / norm
            r_Hx = r_Hx / norm
            r_Hy = r_Hy / norm

            # Compute overlap
            overlap = self.overlap(Ex, Ey, Hx, Hy, r_Ex, r_Ey, r_Hx, r_Hy, x, y)
        else:
            overlap = 0

        # Compute autogradient
        f_x = 2 * np.pi * self.eme.wavelength * (np.conj(overlap))
        power = np.abs(overlap) ** 2

        return f_x, power

    def set_adjoint_sources(self, f_x: complex = 0 + 0j):
        """Computes and places the adjoint sources for use in the adjoint formulation"""
        return [Source(z=2.75, mode_coeffs=[f_x], k=-1)]

    def adjoint_run(self, sources: list):
        """Performs the adjoint run for use in the adjoint formulation"""

        a_grid_x, a_grid_y, a_grid_z, a_field, adjoint_monitor = self.adjoint_results
        a_field *= sources[0].mode_coeffs[0]
        adjoint_monitor.field[:-1] *= sources[0].mode_coeffs[0]
        return a_grid_x, a_grid_y, a_grid_z, a_field, adjoint_monitor

    def optimize(self, design: list, dp=1e-10) -> "np.ndarray":
        """Runs a single step of shape optimization"""

        if self.eme.am_master():
            print("Now performing an optimization...")

        # Update the design region
        design = design if not isinstance(design, np.ndarray) else design.tolist()
        self.set_design(design)

        # Compute the forward run
        grid_x, grid_y, grid_z, X, monitor_forward, monitor_fom = self.forward_run()

        if self.eme.am_master():
            plt.figure()
            monitor_forward.visualize(axes="xz")
            plt.savefig("forward")

        # Compute the partial gradient of the objective function f_x
        f_x, fom = self.objective_gradient(monitor_fom)

        # Calculate the adjoint sources
        sources = self.set_adjoint_sources(f_x)

        # Compute the adjoint run
        grid_x, grid_y, grid_z, lamdagger, monitor_adjoint = self.adjoint_run(sources)

        if self.eme.am_master():
            plt.figure()
            monitor_adjoint.visualize(axes="xz")
            plt.savefig("adjoint")

        # Compute the gradient of the constraint A_u
        A_u = self.gradient(
            grid_x, grid_y, grid_z, dp=dp
        )  # A will need to be 3x3xthe rest to incorporate the right dimensions and cross dielectric etc.

        # Calculate the full gradient of the objective function f_u
        f_u = self.compute_final_gradient(lamdagger, A_u, X)

        plt.figure()
        self.plot_gradients(f_u, monitor_adjoint)
        if self.eme.am_master():
            plt.savefig("gradients")

        # Return the gradient
        return fom, f_u, monitor_forward

    def compute_final_gradient(self, lamdagger: "np.ndarray", A_u: "np.ndarray", X: "np.ndarray"):
        """Computes the final gradient using the adjoint formulation and loops to conserve memory"""

        # Initialize final result
        f_u = np.zeros(A_u.shape[-1], dtype=float)

        # Reshape
        lamdagger = np.transpose(np.conj(lamdagger))
        A_u = A_u
        X = X

        # Loop through all params
        for p in range(len(f_u)):
            A_u_temp = A_u[..., p]

            # Compute all 9 components of the matrix
            A_u_x = np.zeros([3] + list(A_u.shape[2:-1]), dtype=complex)
            for i, mi in enumerate(A_u_temp):
                for j, mij in enumerate(mi):
                    A_u_x[i] += mij * X[j]

            # Compute lambda * A_u_x
            for i in range(3):
                f_u[p] += -2 * np.real(np.sum(A_u_x[i] * lamdagger[..., i].T))

        return f_u

    def draw(self) -> None:
        self.start()
        self.eme.draw()

    def plot_gradients(self, gradients, monitor) -> None:

        # Get the gradients
        design_x, design_z = self.get_design_readable()
        z_start, z_end = self.get_design_region()
        vertices_gradients = np.array([[z, x] for z, x in zip(gradients[1::2], gradients[::2])])
        vertices_origins = np.array([[z + z_start, x] for z, x in zip(design_z, design_x)]).T

        # Get n
        grid_x, grid_z, n_original = monitor.get_array(axes="xz", component="n")
        n_grid_z = np.array([i for i in grid_z if i >= z_start and i <= z_end])
        n = self.get_n(grid_x, None, n_grid_z)

        # Plot gradients
        plt.imshow(np.real(n_original[::-1]), extent=[grid_z[0], grid_z[-1], grid_x[0], grid_x[-1]], cmap="Greys")
        plt.imshow(np.real(n[::-1]), extent=[z_start, z_end, grid_x[0], grid_x[-1]], cmap="Greys")
        plt.quiver(*vertices_origins, vertices_gradients[:, 0], vertices_gradients[:, 1], color="r")

    def calculate_fd_gradient(
        self,
        num_gradients: int = 1,
        dp: float = 1e-4,
        rand: "np.random.RandomState" = None,
        idx: list = None,
        design: list = None,
    ):
        """
        Estimate central difference gradients.

        Parameters
        ----------
        num_gradients : int
            number of gradients to estimate. Randomly sampled from parameters.
        du : float
            finite difference step size

        Returns
        -----------
        fd_gradient : lists
            [number of objective functions][number of gradients]

        """

        # Get the design
        design = self.get_design() if design is None else design
        if num_gradients > len(design):
            raise ValueError(
                "The requested number of gradients must be less than or equal to the total number of design parameters."
            )

        # cleanup
        self.start()

        # preallocate result vector
        fd_gradient = []

        # randomly choose indices to loop estimate
        if idx is None:
            fd_gradient_idx = np.random.choice(
                len(design) // 2,
                num_gradients,
                replace=False,
            ) if rand is None else rand.choice(
                len(design) // 2,
                num_gradients,
                replace=False,
            )
        else:
            fd_gradient_idx = idx[:]

        # loop over indices
        for k in fd_gradient_idx:

            # get current design region
            b0 = np.ones(len(design))
            b0[:] = design[:]

            # assign new design vector
            b0[k * 2] -= dp
            self.set_design(b0)
            self.start()

            # FOM monitor
            source = Source(z=0.25, mode_coeffs=[1], k=1)  # Hard coded
            fom_monitor = self.eme.add_monitor(axes="xy", location=2.75, sources=[source])

            # Adjoint source and monitor
            self.eme.propagate()

            # record final objective function value
            _, fm = self.objective_gradient(fom_monitor)

            # assign new design vector
            b0[k * 2] += 2 * dp
            self.set_design(b0)
            self.start()

            # propagate
            source = Source(z=0.25, mode_coeffs=[1], k=1)  # Hard coded
            fom_monitor = self.eme.add_monitor(axes="xy", location=2.75, sources=[source])
            self.eme.propagate()

            # record final objective function value
            _, fp = self.objective_gradient(fom_monitor)

            # revert design
            b0[k * 2] -= dp
            self.set_design(b0)

            # derivative
            # if self.eme.am_master():
            #     print("end",fp, fm, dp, k, b0[k*2])
            fd_gradient.append((fp - fm) / (2 * dp))

        return fd_gradient, fd_gradient_idx, (fp + fm) / 2, fom_monitor



import os
import numpy as np
import pickle

import torch
import torch.nn as nn

from sklearn.preprocessing import PolynomialFeatures
import sklearn

from emepy.mode import Mode, EigenMode
from emepy.fd import ModeSolver
from emepy.tools import from_chunks, get_epsfunc
from emepy.materials import Si, SiO2

FIELD_WIDTH = 128
FIELD_SIZE = FIELD_WIDTH ** 2


def getUpConvLayer(i_size: int, o_size: int, kernal: int, channels: int, first: bool = False, last: bool = False):
    """Returns the right size for up convolutional sampling"""

    def out_size(in_size, kernal, stride, padding, output_padding):
        return (in_size - 1) * stride - 2 * padding + kernal + output_padding

    stride = 0
    padding = 0
    output_padding = 0

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) < o_size:
            stride += 1

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) > o_size:
            padding += 1

    if not out_size(i_size, kernal, stride, padding, output_padding) == o_size:
        while out_size(i_size, kernal, stride, padding, output_padding) < o_size:
            output_padding += 1

    if first:
        return nn.ConvTranspose2d(
            1, channels, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
        )

    if last:
        return nn.ConvTranspose2d(
            channels, 1, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
        )

    return nn.ConvTranspose2d(
        channels, channels, stride=stride, padding=padding, kernel_size=kernal, output_padding=output_padding
    )


def getDownConvLayer(i_size: int, o_size: int, kernal: int, channels: int, first: bool = False, last: bool = False):
    """Returns the right size for down convolutional sampling"""

    def out_size(in_size, kernal, stride, padding, dilation):
        return (in_size + 2 * padding - dilation * (kernal - 1) - 1) / stride + 1

    stride = 1
    padding = 0
    dilation = 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        while out_size(i_size, kernal, stride, padding, dilation) > o_size:
            dilation += 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        while out_size(i_size, kernal, stride, padding, dilation) < o_size:
            padding += 1

    if not out_size(i_size, kernal, stride, padding, dilation) == o_size:
        raise Exception("Choose a different kernal size")

    if first:
        return nn.Conv2d(1, channels, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)

    if last:
        return nn.Conv2d(channels, 1, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)

    return nn.Conv2d(channels, channels, stride=stride, padding=padding, kernel_size=kernal, dilation=dilation)


class Network(nn.Module):
    """The pytorch inherited class that defines and represents the physical neural network"""

    def __init__(self, code_size: int, channels: int, component: str) -> None:
        """Network constructor

        Parameters
        ----------
        code_size : int
            the number of inputs
        channels : int
            the number of channels
        component : string
            "Hx" or "Hy"
        """
        super().__init__()
        self.channels = channels

        self.linear_up_1 = nn.Linear(code_size, int(FIELD_WIDTH / 20) ** 2)
        self.linear_up_2 = nn.Linear(int(FIELD_WIDTH / 20) ** 2, int(FIELD_WIDTH / 7) ** 2)
        self.linear_up_3 = nn.Linear(int(FIELD_WIDTH / 7) ** 2, int(FIELD_WIDTH / 4) ** 2)
        self.linear_up_4 = nn.Linear(int(FIELD_WIDTH / 4) ** 2, int(FIELD_WIDTH / 3) ** 2)
        self.conv_up_1 = getUpConvLayer(int(FIELD_WIDTH / 3), int(FIELD_WIDTH / 2), 3, channels, first=True)
        self.conv_up_2 = getUpConvLayer(int(FIELD_WIDTH / 2), int(5 * FIELD_WIDTH / 8), 5, channels)
        self.conv_up_3 = getUpConvLayer(int(5 * FIELD_WIDTH / 8), int(6 * FIELD_WIDTH / 8), 7, channels)
        self.conv_up_4 = getUpConvLayer(int(6 * FIELD_WIDTH / 8), int(7 * FIELD_WIDTH / 8), 7, channels)
        self.conv_up_5 = getUpConvLayer(int(7 * FIELD_WIDTH / 8), int(FIELD_WIDTH), 9, channels, last=True)
        self.linear_up_5 = nn.Linear(FIELD_WIDTH ** 2, FIELD_WIDTH ** 2)

        self.relu = nn.ReLU()
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()

        torch.nn.init.xavier_normal_(self.linear_up_1.weight)
        torch.nn.init.xavier_normal_(self.linear_up_2.weight)
        torch.nn.init.xavier_normal_(self.linear_up_3.weight)

        torch.nn.init.xavier_normal_(self.conv_up_1.weight)
        torch.nn.init.xavier_normal_(self.conv_up_2.weight)
        torch.nn.init.xavier_normal_(self.conv_up_3.weight)

        self.component = component

    def forward(self, field) -> tuple:
        """Performs the network propagation

        Parameters
        ----------
        field : array
            The inputs to the network

        Returns
        -------
        tuple (torch array, torch array)
            Returns a tuple of the outputs, inputs
        """

        out = self.tanh(self.linear_up_1(field)).view(-1, 1, int(FIELD_WIDTH / 20) ** 2)
        out = self.tanh(self.linear_up_2(out)).view(-1, 1, int(FIELD_WIDTH / 7) ** 2)
        out = self.tanh(self.linear_up_3(out)).view(-1, 1, int(FIELD_WIDTH / 4) ** 2)
        out = self.tanh(self.linear_up_4(out)).view(-1, 1, int(FIELD_WIDTH / 3), int(FIELD_WIDTH / 3))
        out = self.tanh(self.conv_up_1(out)).view(-1, self.channels, int(FIELD_WIDTH / 2), int(FIELD_WIDTH / 2))
        out = self.tanh(self.conv_up_2(out)).view(-1, self.channels, int(5 * FIELD_WIDTH / 8), int(5 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_3(out)).view(-1, self.channels, int(6 * FIELD_WIDTH / 8), int(6 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_4(out)).view(-1, self.channels, int(7 * FIELD_WIDTH / 8), int(7 * FIELD_WIDTH / 8))
        out = self.tanh(self.conv_up_5(out)).view(-1, 1, FIELD_WIDTH ** 2)

        out = self.linear_up_5(out).view(-1, FIELD_WIDTH, FIELD_WIDTH)

        out = out / 1000.0 if self.component == "Hx" else out / 100.0

        return out, field


class ANN(object):
    """Object that loads the neural network; Users are heavily encouraged to design their own networks and rewrite their own ANN to match their needs"""

    def __init__(self,) -> None:
        """Constructor for Mode Object"""
        self.x = np.linspace(0, 2.5, FIELD_WIDTH)
        self.y = np.linspace(0, 2.5, FIELD_WIDTH)

        self.Hx_model = self.Hx_network()
        self.Hy_model = self.Hy_network()
        self.neff_model = self.neff_regression()

    def neff_regression(self) -> "sklearn.linear_model._base.LinearRegression":
        """Return the opened regression model for the effective index"""

        with open(os.path.dirname(os.path.abspath(__file__)) + "/models/neff_pickle/model.pk", "rb") as f:
            model = pickle.load(f)

        return model

    def Hx_network(self) -> Network:
        """Return the opened network model for the Hx component"""

        from_chunks(os.path.dirname(os.path.abspath(__file__)) + "/models/Hx_chunks/", "hx_temp.pt")
        with open("hx_temp.pt", "rb") as f:
            model = Network(3, 1, "Hx")

            # original saved file with DataParallel
            state_dict = torch.load(f)
            model.load_state_dict(state_dict)

            model.eval()

        os.system("rm hx_temp.pt")
        return model

    def Hy_network(self) -> Network:
        """Return the opened network model for the Hy component"""

        from_chunks(os.path.dirname(os.path.abspath(__file__)) + "/models/Hy_chunks/", "hy_temp.pt")
        with open("hy_temp.pt", "rb") as f:
            model = Network(3, 1, "Hy")

            # original saved file with DataParallel
            state_dict = torch.load(f)
            model.load_state_dict(state_dict)

            model.eval()

        os.system("rm hy_temp.pt")
        return model


class MSNeuralNetwork(ModeSolver):
    """ModeSolver object for the sample neural networks, parameterizes the cross section components. Currently designed only for single mode calculations in Silicon on SiO2"""

    def __init__(self, ann: ANN, wl: float, width: float, thickness: float) -> None:
        """MSNeuralNetwork constructor

        Parameters
        ----------
        ann : ANN
            The ANN object that contains the network and regression models
        wl : number
            The wavelength (most accurate around 1.55 µm)
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        """

        self.wl = wl
        self.width = width
        self.thickness = thickness
        self.ann = ann
        self.Hx_model = ann.Hx_model
        self.Hy_model = ann.Hy_model
        self.neff_model = ann.neff_model
        self.num_modes = 1
        self.x = ann.x
        self.y = ann.y
        self.after_x = self.x
        self.after_y = self.y
        self.mesh = len(self.x) - 1
        self.PML = False
        self.n = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl), compute=True)(
            self.x, self.y
        )

    def solve(self) -> None:
        """Solves for the eigenmode using the neural networks"""

        self.mode = None

        Hx, Hy, neff = self.data(self.width, self.thickness, self.wl)
        self.mode = (Hx, Hy, neff)

    def data(self, width: float, thickness: float, wl: float) -> tuple:
        """Propagates the inputs into the neural networks and regression models and returns the outputs

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)

        Returns
        -------
        tuple (numpy array, numpy array, number)
            Returns Hx, Hy, and neff
        """

        neff = self.neff_regression(width, thickness, wl, self.neff_model)
        Hx = self.Hx_network(width, thickness, wl, self.Hx_model)
        Hy = self.Hy_network(width, thickness, wl, self.Hy_model)

        return Hx, Hy, neff

    def neff_regression(
        self, width: float, thickness: float, wl: float, model: "sklearn.linear_model._base.LinearRegression"
    ) -> float:
        """Calculates the effective index using a regression model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : sklearn regression model
            The model that performs the regression

        Returns
        -------
        number
            Returns the effective index
        """

        poly = PolynomialFeatures(degree=8)
        X = poly.fit_transform([[width, thickness, wl]])
        neff = model.predict(X)

        return neff[0]

    def Hx_network(self, width: float, thickness: float, wl: float, model: Network) -> "np.ndarray":
        """Calculates the Hx component using a network model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : pytorch model
            The model that performs the ann calculation

        Returns
        -------
        numpy array
            Returns the Hx field
        """

        with torch.no_grad():
            parameters = torch.Tensor([[[width, thickness, wl]]])
            output, _ = model(parameters)
            output = output.numpy()
            output = output.reshape(128, 128)

        return output

    def Hy_network(self, width: float, thickness: float, wl: float, model: Network) -> "np.ndarray":
        """Calculates the Hy component using a network model

        Parameters
        ----------
        width : number
            The width of the cross section (most accurate around 550 nm)
        thickness : number
            The thickness of the cross section (most accurate around 250 nm)
        wl : number
            The wavelength (most accurate around 1.55 µm)
        model : pytorch model
            The model that performs the ann calculation

        Returns
        -------
        numpy array
            Returns the Hy field
        """

        with torch.no_grad():
            parameters = torch.Tensor([[[width, thickness, wl]]])
            output, _ = model(parameters)
            output = output.numpy()
            output = output.reshape(128, 128)

        return output

    def clear(self) -> None:
        """Clears the mode in the object"""
        self.mode = None

    def get_mode(self, mode_num: int = 0) -> EigenMode:
        """Returns the solved eigenmode

        Parameters
        ----------
        mode_num : int
            mode index to return mode of

        Returns
        -------
        EigenMode
            the EigenMode corresponding to the provdided mode index
        """
        Hx, Hy, neff = self.mode

        epsfunc_before = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl), compute=True)
        epsfunc_after = get_epsfunc(self.width, self.thickness, 2.5, 2.5, Si(self.wl), SiO2(self.wl))
        m = Mode(
            self.x,
            self.y,
            self.wl,
            neff,
            Hx + 0j,
            Hy + 0j,
            None,
            None,
            None,
            None,
            np.sqrt(epsfunc_after(self.x, self.y)),
        )
        m.compute_other_fields(epsfunc_before, epsfunc_after)
        m.normalize()

        return m



import numpy as np
from matplotlib import pyplot as plt
import pickle
import random
from scipy.signal import convolve2d, convolve
from copy import deepcopy
from typing import Callable

from emepy.tools import compute_other_fields_2D, interp


class EigenMode(object):
    """Virtual class representing an eigenmode"""

    def __init__(self):
        return NotImplementedError()

    def _inner_product(self):
        return NotImplementedError()

    def plot(self):
        return NotImplementedError()

    def get_confined_power(self):
        return NotImplementedError()

    def zero_phase(self):
        return NotImplementedError()

    def plot_material(self):
        return NotImplementedError()

    def compute_other_fields(self):
        return NotImplementedError()

    def save(self, path=("./ModeObject_" + str(random.random()) + ".pk")):
        """Serializes the mode into a pickle file"""
        pickle.dump(self, open(path, "wb+"))

    def get_fields(self):
        """Returns an array [self.Hx, self.Hy, self.Hz, self.Ex, self.Ey, self.Ez]."""

        return [self.Hx, self.Hy, self.Hz, self.Ex, self.Ey, self.Ez]

    def get_H(self):
        """Returns an array [self.Hx, self.Hy, self.Hz]."""

        return [self.Hx, self.Hy, self.Hz]

    def get_E(self):
        """Returns an array [self.Ex, self.Ey, self.Ez]."""

        return [self.Ex, self.Ey, self.Ez]

    def get_neff(self):
        """Returns the effective index as a complex number."""

        return self.neff

    def get_Hx(self):
        return self.Hx

    def get_Hy(self):
        return self.Hy

    def get_Hz(self):
        return self.Hz

    def get_Ex(self):
        return self.Ex

    def get_Ey(self):
        return self.Ey

    def get_Ez(self):
        return self.Ez

    def get_wavelength(self):
        """Returns the wavelength."""

        return self.wl

    def __str__(self):
        return "Mode Object with effective index of " + str(self.neff)

    def change_fields(self, start, other, func):
        for comp in ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]:
            cur = getattr(start, comp)
            new = getattr(other, comp) if isinstance(other, EigenMode) else other
            setattr(start, comp, func(cur, new))
        return start

    def __mul__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a * b)

    def __add__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a + b)

    def __truediv__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a / b)

    def __sub__(self, other):
        return self.change_fields(deepcopy(self), other, lambda a, b: a - b)

    def __imul__(self, other):
        return self.change_fields(self, other, lambda a, b: a * b)

    def __iadd__(self, other):
        return self.change_fields(self, other, lambda a, b: a + b)

    def __itruediv__(self, other):
        return self.change_fields(self, other, lambda a, b: a / b)

    def __isub__(self, other):
        return self.change_fields(self, other, lambda a, b: a - b)

    def inner_product(self, mode2) -> float:
        """Takes the inner product between self and the provided Mode

        Parameters
        ----------
        mode2 : EigenMode
            second eigenmode in the operation

        Returns
        -------
        number
            the inner product between the modes
        """

        return self._inner_product(self, mode2)

    def check_spurious(
        self, threshold_power: float = 0.05, threshold_neff: float = 0.9
    ) -> bool:
        """Takes in a mode and determine whether the mode is likely spurious based on the ratio of confined to not confined power

        Parameters
        ----------
        threshold_power : float
            threshold of power percentage of Pz in the core to total
        threshold_neff : float
            threshold of real to abs neff percentage

        Returns
        -------
        boolean
            True if likely spurious
        """

        power_bool = self.get_confined_power() < threshold_power
        neff_bool = (np.real(self.neff) / np.abs(self.neff)) < threshold_neff
        return power_bool or neff_bool

    def normalize(self) -> None:
        """Normalizes the Mode to power 1."""
        self.zero_phase()
        factor = self.inner_product(self)
        self /= np.sqrt(factor)


class Mode1D(EigenMode):
    """Object that holds the field profiles and effective index for a 1D eigenmode"""

    def __init__(
        self,
        x: "np.ndarray" = None,
        wl: float = None,
        neff: float = None,
        Hx: "np.ndarray" = None,
        Hy: "np.ndarray" = None,
        Hz: "np.ndarray" = None,
        Ex: "np.ndarray" = None,
        Ey: "np.ndarray" = None,
        Ez: "np.ndarray" = None,
        n: "np.ndarray" = None,
    ) -> None:
        """Constructor for Mode1D Object (one dimensional eigenmode)

        Parameters
        ----------
        x : (ndarray float)
            array of grid points in x direction (propogation in z)
        wl : (float)
            wavelength (meters)
        neff : (float)
            effective index
        Hx : (ndarray float)
            Hx field profile
        Hy : (ndarray float)
            Hy field profile
        Hz : (ndarray float)
            Hz field profile
        Ex : (ndarray float)
            Ex field profile
        Ey : (ndarray float)
            Ey field profile
        Ez : (ndarray float)
            Ez field profile
        n : (ndarray float)
            refractive index profile
        """

        self.x = x
        self.wl = wl
        self.neff = neff
        self.Hx = Hx if Hx is not None else np.zeros(10)
        self.Hy = Hy if Hy is not None else self.Hx * 0
        self.Hz = Hz if Hz is not None else self.Hx * 0
        self.Ex = Ex if Ex is not None else self.Hx * 0
        self.Ey = Ey if Ey is not None else self.Hx * 0
        self.Ez = Ez if Ez is not None else self.Hx * 0
        self.n = n
        self.H = np.sqrt(
            np.abs(self.Hx) ** 2 + np.abs(self.Hy) ** 2 + np.abs(self.Hz) ** 2
        )
        self.E = np.sqrt(
            np.abs(self.Ex) ** 2 + np.abs(self.Ey) ** 2 + np.abs(self.Ez) ** 2
        )

    def plot(self, operation: str = "Real", normalize: bool = True) -> None:
        """Plots the fields in the mode using pyplot. Should call plt.figure() before and plt.show() or plt.savefig() after

        Parameters
        ----------
        operation : string or function
            the operation to perform on the fields from ("Real", "Imaginary", "Abs", "Abs^2") (default:"Real") or a function such as np.abs
        normalize : bool
            if true, will normalize biggest field to 1
        """

        temp = (
            self
            / max(
                [
                    np.abs(np.real(np.amax(i)))
                    for i in [self.Ex, self.Ey, self.Ez, self.Hx, self.Hy, self.Hz]
                ]
            )
            if normalize
            else self / 1
        )

        # Parse operation
        op_name = (
            operation.__name__ if hasattr(operation, "__name__") else str(operation)
        )
        if operation == "Imaginary":
            operation = lambda a: np.imag(a)
        elif operation == "Abs":
            operation = lambda a: np.abs(a)
        elif operation == "Abs^2":
            operation = lambda a: np.abs(a) ** 2
        elif operation == "Real":
            operation = lambda a: np.real(a)
        try:
            t = self.change_fields(
                deepcopy(temp), deepcopy(temp), lambda a, b: operation(b)
            )
            Hx, Hy, Hz, Ex, Ey, Ez = [t.Hx, t.Hy, t.Hz, t.Ex, t.Ey, t.Ez]
        except Exception as e:
            print(e)
            raise Exception(
                "Invalid operation provided. Please choose from ('Imaginary', 'Abs', 'Abs^2', 'Real') or provide a function"
            )

        # Plot fields
        fields = ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]
        for i, field in enumerate([Hx, Hy, Hz, Ex, Ey, Ez]):
            plt.subplot(2, 3, i + 1)
            plt.plot(self.x, field)
            plt.xlabel("x µm")
            plt.ylabel("{}({})".format(op_name, fields[i]))
        plt.tight_layout()

    def _inner_product(
        self, mode1: EigenMode, mode2: EigenMode, mask: "np.ndarray" = None
    ) -> float:
        """Helper function that takes the inner product between Modes mode1 and mode2

        Parameters
        ----------
        mode1 : Mode
            first eigenmode in the operation
        mode2 : Mode
            second eigenmode in the operation
        mask : np.ndarray
            a mask to multiply on the field profiles before conducting the inner product

        Returns
        -------
        number
            the inner product between the two input modes
        """

        mask = 1 if mask is None else mask

        Ex = mode1.Ex * mask
        Hy = np.conj(mode2.Hy) * mask
        Ey = mode1.Ey * mask
        Hx = np.conj(mode2.Hx) * mask

        cross = Ex * Hy - Ey * Hx

        return np.trapz(cross, np.real(mode1.x))

    def get_confined_power(self, num_pixels: int = None) -> float:
        """Takes in a mode and returns the percentage of power confined in the core

        Parameters
        ----------
        num_pixels : int
            number of pixels outside of the core to expand the mask to capture power just outside the core

        Returns
        -------
        float
            Percentage of confined power
        """

        # Increase core by 5% to capture slight leaks
        if num_pixels is None:
            num_pixels = int(len(self.x) * 0.05)

        mask = np.where(self.n > np.mean(self.n), 1, 0)
        kernel = np.ones(num_pixels + 1)
        mask = convolve(mask, kernel, "same")
        mask = np.where(mask > 0, 1, 0)
        ratio = self._inner_product(self, self, mask=mask) / self._inner_product(
            self, self, mask=None
        )
        return ratio

    def zero_phase(self) -> None:
        """Changes the phase such that the z components are all imaginary and the xy components are all real."""

        index = int(self.Hy.shape[0] / 2)
        phase = np.angle(np.array(self.Hy))[index]
        self *= np.exp(-1j * phase)
        if (np.sum(np.real(self.Hy))) < 0:
            self *= -1

    def plot_material(
        self, operation: Callable[["np.ndarray"], "np.ndarray"] = np.real
    ) -> None:
        """Plots the index of refraction profile"""
        plt.plot(self.x, operation(self.n))
        plt.title("Index of Refraction")
        plt.xlabel("x (µm)")
        plt.ylabel("y (µm)")


class Mode(EigenMode):
    """Object that holds the field profiles and effective index for a 2D eigenmode"""

    def __init__(
        self,
        x: "np.ndarray" = None,
        y: "np.ndarray" = None,
        wl: float = None,
        neff: float = None,
        Hx: "np.ndarray" = None,
        Hy: "np.ndarray" = None,
        Hz: "np.ndarray" = None,
        Ex: "np.ndarray" = None,
        Ey: "np.ndarray" = None,
        Ez: "np.ndarray" = None,
        n: "np.ndarray" = None,
    ) -> None:
        """Constructor for Mode Object

        Parameters
        ----------
        x : (ndarray float)
            array of grid points in x direction (propogation in z)
        y : (ndarray float)
            array of grid points in y direction (propogation in z)
        wl : (float)
            wavelength (meters)
        neff : (float)
            effective index
        Hx : (ndarray float)
            Hx field profile
        Hy : (ndarray float)
            Hy field profile
        Hz : (ndarray float)
            Hz field profile
        Ex : (ndarray float)
            Ex field profile
        Ey : (ndarray float)
            Ey field profile
        Ez : (ndarray float)
            Ez field profile
        n : (ndarray float)
            refractive index profile
        """

        self.x = x
        self.y = y
        self.wl = wl
        self.neff = neff
        self.Hx = Hx if Hx is not None else np.zeros((10, 10))
        self.Hy = Hy if Hy is not None else self.Hx * 0
        self.Hz = Hz if Hz is not None else self.Hx * 0
        self.Ex = Ex if Ex is not None else self.Hx * 0
        self.Ey = Ey if Ey is not None else self.Hx * 0
        self.Ez = Ez if Ez is not None else self.Hx * 0
        self.n = n
        self.H = np.sqrt(
            np.abs(self.Hx) ** 2 + np.abs(self.Hy) ** 2 + np.abs(self.Hz) ** 2
        )
        self.E = np.sqrt(
            np.abs(self.Ex) ** 2 + np.abs(self.Ey) ** 2 + np.abs(self.Ez) ** 2
        )

    def plot(
        self, operation: str = "Real", colorbar: bool = True, normalize: bool = True
    ) -> None:
        """Plots the fields in the mode using pyplot. Should call plt.figure() before and plt.show() or plt.savefig() after

        Parameters
        ----------
        operation : string or function
            the operation to perform on the fields from ("Real", "Imaginary", "Abs", "Abs^2") (default:"Real") or a function such as np.abs
        colorbar : bool
            if true, will show a colorbar for each field
        normalize : bool
            if true, will normalize biggest field to 1
        """

        temp = (
            self
            / max(
                [
                    np.abs(np.real(np.amax(i)))
                    for i in [self.Ex, self.Ey, self.Ez, self.Hx, self.Hy, self.Hz]
                ]
            )
            if normalize
            else self / 1
        )

        # Parse operation
        op_name = (
            operation.__name__ if hasattr(operation, "__name__") else str(operation)
        )
        if operation == "Imaginary":
            operation = lambda a: np.imag(a)
        elif operation == "Abs":
            operation = lambda a: np.abs(a)
        elif operation == "Abs^2":
            operation = lambda a: np.abs(a) ** 2
        elif operation == "Real":
            operation = lambda a: np.real(a)
        try:
            t = self.change_fields(
                deepcopy(temp), deepcopy(temp), lambda a, b: operation(b)
            )
            Hx, Hy, Hz, Ex, Ey, Ez = [t.Hx, t.Hy, t.Hz, t.Ex, t.Ey, t.Ez]
        except Exception as e:
            print(e)
            raise Exception(
                "Invalid operation provided. Please choose from ('Imaginary', 'Abs', 'Abs^2', 'Real') or provide a function"
            )

        # Plot fields
        fields = ["Hx", "Hy", "Hz", "Ex", "Ey", "Ez"]
        for i, field in enumerate([Hx, Hy, Hz, Ex, Ey, Ez]):
            plt.subplot(
                2, 3, i + 1, adjustable="box", aspect=field.shape[0] / field.shape[1]
            )
            v = max(abs(field.min()), abs(field.max()))
            plt.imshow(
                np.rot90(field),
                cmap="RdBu",
                vmin=-v,
                vmax=v,
                extent=[
                    temp.x[0],
                    temp.x[-1],
                    temp.y[0],
                    temp.y[-1],
                ],
                interpolation="none",
            )
            plt.title("{}({})".format(op_name, fields[i]))
            if colorbar:
                plt.colorbar(fraction=0.046, pad=0.04)
            plt.xlabel("x µm")
            plt.ylabel("y µm")

        plt.tight_layout()

    def _inner_product(
        self, mode1: EigenMode, mode2: EigenMode, mask: "np.ndarray" = None
    ) -> float:
        """Helper function that takes the inner product between Modes mode1 and mode2

        Parameters
        ----------
        mode1 : EigenMode
            first eigenmode in the operation
        mode2 : EigenMode
            second eigenmode in the operation
        mask : np.ndarray
            a mask to multiply on the field profiles before conducting the inner product
        Returns
        -------
        number
            the inner product between the two input modes
        """

        mask = 1 if mask is None else mask

        Ex = mode1.Ex * mask
        Hy = np.conj(mode2.Hy) * mask
        Ey = mode1.Ey * mask
        Hx = np.conj(mode2.Hx) * mask

        cross = Ex * Hy - Ey * Hx

        return np.trapz(np.trapz(cross, np.real(mode1.x)), np.real(mode1.y))

    def get_confined_power(self, num_pixels: int = None) -> float:
        """Takes in a mode and returns the percentage of power confined in the core

        Parameters
        ----------
        num_pixels : int
            number of pixels outside of the core to expand the mask to capture power just outside the core

        Returns
        -------
        float
            Percentage of confined power
        """

        # Increase core by 5% to capture slight leaks
        if num_pixels is None:
            num_pixels = int(len(self.x) * 0.05)

        mask = np.where(self.n > np.mean(self.n), 1, 0)
        kernel = np.ones((num_pixels + 1, num_pixels + 1))
        mask = convolve2d(mask, kernel, "same")
        mask = np.where(mask > 0, 1, 0)
        ratio = self._inner_product(self, self, mask=mask) / self._inner_product(
            self, self, mask=None
        )
        return ratio

    def zero_phase(self) -> None:
        """Changes the phase such that the z components are all imaginary and the xy components are all real."""

        index = int(self.Hy.shape[0] / 2)
        phase = np.angle(np.array(self.Hy))[index][index]
        self *= np.exp(-1j * phase)
        if (np.sum(np.real(self.Hy))) < 0:
            self *= -1

    def plot_material(self) -> None:
        """Plots the index of refraction profile"""

        plt.figure()
        plt.imshow(
            np.real(np.rot90(self.n)),
            extent=[
                self.x[0],
                self.x[-1],
                self.y[0],
                self.y[-1],
            ],
            cmap="Greys",
            interpolation="none",
        )
        plt.colorbar()
        plt.title("Index of Refraction")
        plt.xlabel("x (µm)")
        plt.ylabel("y (µm)")

    def compute_other_fields(
        self,
        epsfunc_1: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
        epsfunc_2: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
        boundary: str = "0000",
    ) -> None:
        """Given the Hx and Hy fields, maxwell's curl relations can be used to calculate the remaining field; adapted from the EMpy

        Parameters
        ----------
        epsfunc_1 : function
            epsfunc for computing other fields
        epsfunc_2 : function
            epsfunc for computing final refractive index profile
        boundary : str
            the boundary conditions as defined by electromagneticpython
        """

        (
            self.Hx,
            self.Hy,
            self.Hz,
            self.Ex,
            self.Ey,
            self.Ez,
        ) = compute_other_fields_2D(
            self.neff, self.Hx, self.Hy, self.wl, self.x, self.y, boundary, epsfunc_1
        )
        x_ = (self.x[1:] + self.x[:-1]) / 2.0
        y_ = (self.y[1:] + self.y[:-1]) / 2.0
        self.Ex = interp(self.x, self.y, x_, y_, self.Ex, False)
        self.Ey = interp(self.x, self.y, x_, y_, self.Ey, False)
        self.Ez = interp(self.x, self.y, x_, y_, self.Ez, False)
        self.x = self.x - self.x[int(len(self.x) / 2)]
        self.y = self.y - self.y[int(len(self.y) / 2)]
        self.H = np.sqrt(
            np.abs(self.Hx) ** 2 + np.abs(self.Hy) ** 2 + np.abs(self.Hz) ** 2
        )
        self.E = np.sqrt(
            np.abs(self.Ex) ** 2 + np.abs(self.Ey) ** 2 + np.abs(self.Ez) ** 2
        )
        self.n = np.sqrt(epsfunc_2(self.x, self.y))



import numpy as np
import numpy
from tqdm import tqdm
from emepy.monitors import Monitor
import matplotlib
import simphony
import pickle

import importlib

if importlib.util.find_spec("mpi4py") is not None:
    from mpi4py import MPI

from emepy.fd import MSEMpy
from emepy.models import Duplicator, Model, ModelTools, Layer, InterfaceSingleMode, InterfaceMultiMode

_prop_all = ModelTools._prop_all
_prop_all_wrapper = ModelTools._prop_all_wrapper
purge_spurious = ModelTools.purge_spurious
get_sources = ModelTools.get_sources
get_source_system = ModelTools.get_source_system
periodic_duplicate_format = ModelTools.periodic_duplicate_format
make_copy_model = ModelTools.make_copy_model
compute = ModelTools.compute
layers_task = ModelTools.layers_task
if importlib.util.find_spec("mpi4py") is not None:
    am_master = lambda parallel=True: MPI.COMM_WORLD.Get_rank() == 0 if parallel else True
else:
    am_master = lambda parallel=True: True


class EME(object):
    """The EME class is the heart of the package. It provides the algorithm that cascades sections modes together to provide the s-parameters for a geometric structure. The object is dependent on the Layer objects that are fed inside."""

    states = {
        0: "start",
        1: "mode_solving",
        2: "finished_modes",
        3: "layer_propagating",
        4: "finished_layer",
        5: "network_building",
        6: "finished_network",
        7: "field_propagating",
        8: "finished",
    }

    def __init__(
        self,
        layers: list = [],
        num_periods: int = 1,
        mesh_z: int = 200,
        parallel: bool = False,
        quiet: bool = False,
        **kwargs
    ) -> None:
        """EME class constructor

        Parameters
        ----------
        layers : list [Layer]
            An list of Layer objects, arranged in the order they belong geometrically. (default: [])
        num_periods : int
            Number of periods if defining a periodic structure (default: 1)
        mesh_z : int
            Number of mesh points in z per period for default monitors (default: 200)
        parallel : bool
            If true, will allocate parallelized processes for solving modes, propagating layers, and filling monitors with field data (default: False)
        quiet : bool
            If true, will not print current state and status of the solver (default: False)
        """

        self.parallel = parallel
        if parallel:
            self._configure_parallel_resources()
        else:
            self.size = 1
        self.quiet = quiet or not self.am_master()
        self.reset(parallel=parallel, configure_parallel=False)
        self.layers = layers[:]
        self.num_periods = num_periods
        self.mesh_z = mesh_z
        self.monitors = []
        self.custom_monitors = []
        self.forward_periodic_s = []
        self.reverse_periodic_s = []
        self.parallel = parallel
        self.network = None

    def add_layer(self, layer: Layer) -> None:
        """The add_layer method will add a Layer object to the EME object. The object will be geometrically added to the very right side of the structure. Using this method after propagate is useless as the solver has already been called.

        Parameters
        ----------
        layer : Layer
            Layer object to be appended to the list of Layers inside the EME object.
        """

        self.layers.append(layer)

    def add_layers(self, *layers) -> None:
        """Calls add layers for the layers provided"""

        for layer in layers:
            self.add_layer(layer)

    def reset(self, full_reset: bool = True, parallel: bool = False, configure_parallel: bool = True) -> None:
        """Clears out the layers and s params so the user can reuse the object in memory on a new geometry

        Parameters
        ----------
        full_reset : boolean
            If true, will reset everything inside of the object and allow for reinstancing without memory issues (default: True)
        parallel : boolean
            If configure_parallel is True, after reset this method will set the value of parallel. Similar to the constructor (default: False)
        configure_parallel : boolean
            If configure_parallel is True, after reset this method will set the value of parallel. Similar to the constructor (default: True)
        """

        # Erase all information except number of periods
        if full_reset:
            self.layers = []
            self.wavelength = None
            self.parallel = parallel
            if parallel and configure_parallel:
                self._configure_parallel_resources()
            self._update_state(0)

        # Only unsolve everything and erase monitors
        else:
            for i in range(len(self.layers)):
                self.layers[i].clear()
            self.activated_layers = []
            self._update_state(0)

        self.monitors = []
        self.custom_monitors = []
        self.networks = None
        self.network = None

    def solve_modes(self) -> None:
        """Solves for the modes in the system and is the first step in the solver's process all in parallel"""

        # Check if already solved
        if self.state > 1:
            return

        # Hold still while solving finishes
        while self.state == 1:
            continue

        # Solve modes
        self._update_state(1)
        tasks = []
        if self.am_master():
            for layer in self.layers:
                tasks.append((layer.begin_activate()[0], [layer.begin_activate()[1]], {}))

            # Organized solved layers
            self.activated_layers = dict(zip(range(self.num_periods), [[] for _ in range(self.num_periods)]))
        results = self._run_parallel_functions(*tasks)
        final_results = []
        if self.am_master() and results is not None:
            sources = self.get_sources()
            length = 0.0
            for mode_solver, layer in zip(results, self.layers):
                solved = layer.finish_activate(
                    sources=sources, start=length, period_length=self._get_total_length(), mode_solver=mode_solver
                )
                length += layer.length
                final_results.append(solved)
            for solved in final_results:
                for per, layers in solved.items():
                    if not sum([i is not None for so in final_results for p, s in so.items() for i in s if p == per]):
                        self.activated_layers[per] = None
                    else:
                        self.activated_layers[per] += layers if layers[0] is not None else solved[0]

        self._update_state(2)

    def propagate_layers(self) -> None:
        """Propagates each layer with the next by creating interface models and cascading all in parallel. This is the second step for the solver"""

        # Check state
        if self.state == 2:
            self._update_state(3)
        else:
            raise Exception("In the wrong place")

        # Only master needs to control the work
        if self.am_master():
            # Create return dict
            final_activated_layers = dict(zip(range(self.num_periods), [None for _ in range(self.num_periods)]))
            periodic_interfaces = dict(zip(range(self.num_periods), [None for _ in range(self.num_periods)]))

            # Loop through all periods in case of custom sources
            for per, activated_layers in self.activated_layers.items():
                tasks = []

                # Most cases
                if per and activated_layers is None:
                    continue

                # See if only one layer or no layer
                if not per and activated_layers is None:
                    raise Exception("No activated layers in system")
                elif len(activated_layers) == 1:
                    periodic_interfaces[per] = InterfaceMultiMode(activated_layers[-1], activated_layers[0])
                    final_activated_layers[per] = activated_layers[0]
                    continue

                # Configure interface
                num_modes = max([len(l.modes) for l in activated_layers])
                interface_type = InterfaceSingleMode if num_modes == 1 else InterfaceMultiMode

                # Setup parallel layer prop tasks
                for left, right in zip(activated_layers[:-1], activated_layers[1:]):
                    tasks.append((layers_task, [left, right, interface_type], {}))

                # Propagate
                results = self._run_parallel_functions(*tasks)

                # Calculate period interface in case multiple periods
                periodic_interfaces[per] = InterfaceMultiMode(activated_layers[-1], activated_layers[0])

                # Only keep what we need
                for i, result in enumerate(results):
                    layer, cascaded = result
                    attributes = layer.__dict__
                    activated_layers[i] = cascaded
                    activated_layers[i].length = attributes["length"]
                    activated_layers[i].wavelength = attributes["wavelength"]
                    activated_layers[i].modes = attributes["modes"]
                    activated_layers[i].num_modes = attributes["num_modes"]

                # Add to cascaded list
                final_activated_layers[per] = _prop_all(*activated_layers)

            # Update
            self.networks = final_activated_layers
            self.periodic_interfaces = periodic_interfaces

        # Ensure the workers meet up to work
        else:
            self._run_parallel_functions()

        # Finish state
        self._update_state(4)

    def build_network(self) -> None:
        """Builds the full network from the cascaded layers. This is the third step in the solving process."""

        # Initialize building state
        self._update_state(5)

        if self.am_master():
            # Place all periods' proper networks into the final network
            networks = []
            for per, network in self.networks.items():
                start = per * self._get_total_length()
                end = (per + 1) * self._get_total_length()
                n = network if network is not None else self.networks[0]
                networks.append(periodic_duplicate_format(n, start, end))
                if not per:
                    networks.append(
                        self.periodic_interfaces[per]
                        if self.periodic_interfaces[per] is not None
                        else self.periodic_interfaces[0]
                    )

            # Propagate final network with all periods
            self.network = _prop_all(*networks)

        else:
            self.network = None

        if self.size > 1:
            self.network = self.comm.scatter([self.network] * self.size, root=0)

        # Finish building state
        self._update_state(6)

    def field_propagate(self, left_coeffs: list, right_coeffs: list) -> None:
        """Propagates the modes through the device to calculate the field profile everywhere

        Parameters
        ----------
        left_coeffs : list
            A list of floats that represent the mode coefficients for the left side of the full geometry
        right_coeffs : list
            A list of floats that represent the mode coefficients for the right side of the full geometry
        """

        # Start state
        self._update_state(7)

        # Each layer earns a process
        if self.am_master():

            # Create tasks
            cur_len = 0
            tasks = []

            # Loop through all layers
            per_li_map = {}
            for per, activated_layers in self.activated_layers.items():
                activated_layers = activated_layers if activated_layers is not None else self.activated_layers[0]
                for i, layer in enumerate(activated_layers):
                    prop, checked_l = self._layer_field_propagate_part1(
                        i * 1, make_copy_model(layer), per * 1, cur_len * 1
                    )
                    tasks.append(
                        (
                            _prop_all_wrapper,
                            [
                                [
                                    make_copy_model(t, False)
                                    for t in prop
                                    if (t is not None) and not (isinstance(t, list) and not len(t))
                                ],
                                checked_l.pins,
                            ],
                            {},
                        )
                    )
                    cur_len += layer.length
                    per_li_map[len(tasks) - 1] = (per, i)

            # Get results
            results = self._run_parallel_functions(*tasks)

            # Update monitors
            cur_len = 0
            for i, result in enumerate(results):
                S, checked_l_pins = result
                per, layer_index = per_li_map[i]
                activated_layers = self.activated_layers[per]
                # layer_index = i % len(self.activated_layers[0])
                # activated_layers = self.activated_layers[i // len(self.activated_layers[0])]
                activated_layers = activated_layers if activated_layers is not None else self.activated_layers[0]

                # Loop through all monitors
                for m in self.monitors + self.custom_monitors:
                    layer = make_copy_model(activated_layers[layer_index])
                    z_list = m.get_z_list(cur_len, cur_len + layer.length)  # ## ## # ## fix this

                    # If the monitor is in the layer
                    if not len(z_list):
                        continue

                    # Get the field profiles
                    just_z_list = [i[1] for i in z_list]
                    z_index = [i[0] for i in z_list]
                    fields = self._layer_field_propagate_part2(
                        layer_index,
                        layer,
                        left_coeffs[:],
                        right_coeffs[:],
                        cur_len * 1,
                        just_z_list[:],
                        m,
                        S,
                        checked_l_pins,
                    )

                    # Update the monitor
                    for i, f in zip(z_index, fields):
                        self._set_monitor(m, i, f)

                # Update current length
                cur_len = cur_len + layer.length

        # Ensure the workers are there to work
        else:
            self._run_parallel_functions()

        # Finish state
        self._update_state(8)

    def am_master(self) -> bool:
        """Returns true for the master process if the user is running a parallel process using mpi. This is essential for I/O"""
        return self.rank == 0 if self.parallel else True

    def get_sources(self) -> dict:
        """Returns a dictionary of each period and the Source objects that can be found inside each"""
        srcs = dict(zip(range(self.num_periods), [[] for _ in range(self.num_periods)]))
        for per in range(self.num_periods):
            start = self._get_total_length() * per
            end = self._get_total_length() * (per + 1)
            src = get_sources([j for i in (self.monitors + self.custom_monitors) for j in i.sources], start, end)
            src.sort(key=lambda s: s.z)
            srcs[per] = src

        return srcs

    def s_parameters(self, freqs=None) -> "numpy.ndarray":
        """Returns the s_parameters if they exist. If they don't exist yet, propagate() will be called first.

        Returns
        -------
        numpy array
            The s_params acquired during propagation
        """

        if self.network is None:
            self.propagate()
        return self.network.s_parameters(freqs)

    def add_monitor(
        self,
        axes: str = "xz",
        sources: list = [],
        mesh_z: int = None,
        z_range: tuple = None,
        location: float = None,
        components: list = None,
        exempt: bool = True,
    ) -> Monitor:
        """Creates a monitor associated with the eme object BEFORE the simulation is ran

        Parameters
        ----------
        axes : str
            the spacial axes to capture fields in. Options : 'xz' (default), 'xy', 'xz', 'xyz', 'x', 'y', 'z'. Currently only 'xz' is implemented. Note, propagation is always in z.
        sources : list[Source]
            the user can specify custom mode sources to use for this monitor (default: input left)
        mesh_z : int
            number of mesh points in z (for periodic structures, will be z * num_periods), warning: if different than global value for EME, a separate run will have to take place for this monitor (default: EME global defined)
        z_range : tuple
            tuple or list of the form (start, end) representing the range of the z values to extract
        location : float
            z coordinate where to save data for a 'xy' monitor
        components : list[string]
            a list of the field components to include. Unless the user is worried about memory, this is best left alone. (Default: ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz", "n"])
        exempt : bool
            flag used for very specific case when using PML for MSEMpy. The user never has to change this value.

        Returns
        -------
        Monitor
            the newly created Monitor object
        """

        # Establish mesh # Current workaround...not final
        components = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz", "n"] if components is None else components
        if exempt and (self.layers[0].mode_solver.PML and isinstance(self.layers[0].mode_solver, MSEMpy)):
            x = (
                len(
                    self.layers[0].mode_solver.x[
                        self.layers[0].mode_solver.nlayers[1] : -self.layers[0].mode_solver.nlayers[0]
                    ]
                )
                + 1
            )
            y = (
                len(
                    self.layers[0].mode_solver.y[
                        self.layers[0].mode_solver.nlayers[3] : -self.layers[0].mode_solver.nlayers[2]
                    ]
                )
                + 1
            )
        else:
            x = len(self.layers[0].mode_solver.after_x)
            y = len(self.layers[0].mode_solver.after_y)

        # # Default Source
        # if not len(sources):
        #     sources.append(Source())

        # Default mesh_z
        mesh_z = mesh_z if mesh_z is not None else self.mesh_z

        # Ensure the axes is not still under development
        if axes in ["xz", "zx", "yz", "zy", "xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:

            # Create lengths
            l = self._get_total_length()
            single_lengths = np.linspace(0, l, mesh_z, endpoint=False).tolist()
            lengths = np.linspace(0, l, mesh_z, endpoint=False).tolist()

            for i in range(1, self.num_periods):
                lengths += (np.array(single_lengths) + l * i).tolist()
            lengths = [lengths for _ in range(len(components))]
            single_lengths = [single_lengths for _ in range(len(components))]

            # Ensure z range is in proper format
            try:
                start, end = [lengths[0][0], lengths[0][-1]] if z_range is None else z_range
            except Exception as e:
                raise Exception(
                    "z_range should be a tuple or list of the form (start, end) "
                    "representing the range of the z values to extract where start"
                    "and end are floats such as (0, 1e-6) for a 1 µm range"
                ) from e

            # Fix z mesh if changed by z_range
            difference_start = lambda list_value: abs(list_value - start)
            difference_end = lambda list_value: abs(list_value - end)
            s = min(np.array(lengths)[0], key=difference_start)
            e = min(np.array(lengths)[0], key=difference_end)
            z = np.sum((s <= np.array(lengths)[0]) * (e >= np.array(lengths)[0]))

            # Create monitor dimensions
            c = len(components)
            if axes in ["xz", "yz"]:
                dimensions = (c, x, z)
            elif axes in ["yz", "zy"]:
                dimensions = (c, y, z)
            elif axes in ["xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
                dimensions = (c, x, y, z)
            elif axes in ["xy", "yx"]:
                dimensions = (c, x, y)
            else:
                raise Exception("Improper axes format")

            # Create grids
            grid_x = self.layers[0].mode_solver.after_x
            grid_y = self.layers[0].mode_solver.after_y
            grid_z = np.linspace(s, e, z)

        elif axes in ["xy", "yx"]:

            # No z_range needed
            z_range = None

            # Create single length
            lengths = [[location] for _ in range(len(components))]
            single_lengths = lengths

            # Create monitor dimensions
            c = len(components)
            dimensions = (c, x, y)

            # Create grids
            grid_x = self.layers[0].mode_solver.x
            grid_y = self.layers[0].mode_solver.y
            grid_z = np.array([location])

        else:
            raise Exception(
                "Monitor setup {} has not yet been implemented. Please choose from the following implemented monitor types: ['xz', 'yz', 'xyz']".format(
                    axes
                )
            )

        # Create monitor
        total_length = self._get_total_length() * self.num_periods
        monitor = Monitor(
            axes,
            dimensions,
            components,
            z_range,
            grid_x,
            grid_y,
            grid_z,
            location,
            sources=sources,
            total_length=total_length,
        )

        # Place monitor where it belongs
        if len(lengths[0]) == self.mesh_z:
            self.monitors.append(monitor)
        else:
            self.custom_monitors.append(monitor)

        return monitor

    def draw(
        self, z_range: tuple = None, mesh_z: int = 200, plot_sources: bool = True, plot_xy_sources=True
    ) -> "matplotlib.image.AxesImage":
        """The draw method sketches a rough approximation for the xz geometry of the structure using pyplot where x is the width of the structure and z is the length. This will change in the future.

        Parameters
        ----------
        z_range : tuple
            tuple or list of the form (start, end) representing the range of the z values to extract
        mesh_z : int
            the number of mesh points in z to calculate index profiles for

        Returns
        -------
        matplotlib.image.AxesImage
            the image used to plot the index profile
        """

        # Setup temp monitor
        sources = [] if not plot_sources else self.get_sources()[0]
        temp_storage = [self.monitors, self.custom_monitors]
        xy_monitors = [i for j in temp_storage for i in j if i.axes in ["xy", "yx"]] if plot_xy_sources else []
        self.monitors, self.custom_monitors = [[], []]
        monitor = self.add_monitor(
            axes="xz", components=["n"], z_range=z_range, exempt=False, mesh_z=mesh_z, sources=sources
        )

        # Fix monitor
        if plot_sources and not len(sources):
            self._assign_monitor_sources([1], [])
        monitor.xy_monitors = xy_monitors

        # Draw
        self._propagate_n_only()
        im = monitor.visualize(component="n", show_xy_monitors=True)
        self.monitors, self.custom_monitors = temp_storage
        return im

    def propagate(self, left_coeffs: list = None, right_coeffs: list = []) -> "simphony.models.Model":
        """The propagate method should be called once all Layer objects have been added. This method will call the EME solver and produce s-parameters. The defulat

        Parameters
        ----------
        left_coeffs : list
            A list of floats that represent the mode coefficients for the left side of the full geometry. The default is determined on whether or not any custom mode sources or right_coeffs are defined. If they are, (default:[]) else (default"[1])
        right_coeffs : list
            A list of floats that represent the mode coefficients for the right side of the full geometry. (default:[])

        Returns
        -------
        simphony.models.Model
            The simphony model that represents the entire device
        """

        # Check for layers
        if not len(self.layers):
            raise Exception("Must place layers before propagating")
        else:
            self.wavelength = self.layers[0].wavelength

        # Fix defaults
        if left_coeffs is None:
            left_coeffs = (
                [1] if not len(right_coeffs) and not len([i for i in self.get_sources().values() if len(i)]) else []
            )
            if len(left_coeffs) and self.am_master():
                print("No sources defined, defaulting to left_coeffs=[1]")

        # Assign monitor sources
        self._assign_monitor_sources(left_coeffs, right_coeffs)

        # Solve for the modes
        if self.state == 0:
            self.solve_modes()

        # Forward pass
        if self.state == 2:
            self.propagate_layers()

        # Periodic ### ToDo, parallelize this process using a log2 n technique
        if self.state == 4:
            self.build_network()

        # Update monitors
        if self.state >= 6:
            self.field_propagate(left_coeffs, right_coeffs)

        return self.network

    def batch_scatter(self, data, root=0, limit=2 ** 30):
        """Scatters data to all workers in a batched manner that will not exceed the MPI integer limit"""

        # Initialize lists
        scattered = []
        serialized_data = [pickle.dumps(i, 0) for i in data] if data is not None else []
        length = sum([len(i) for i in serialized_data])
        num_divisors = self.comm.scatter([length // limit + 1] * self.size, root=root)

        # Wait
        self.comm.barrier()

        # Loop through every divisor
        for j in range(num_divisors):

            # New divisor to scatter
            to_scatter = []
            for i, serial in enumerate(serialized_data):  # One for every worker
                start = j * len(serial) // num_divisors
                end = (
                    (j + 1) * len(serial) // num_divisors
                    if not j == num_divisors - 1
                    else (j + 2) * len(serial) // num_divisors
                )
                ser = serial[start:end]
                to_scatter.append(ser)

            # Distribute
            scattered.append(self.comm.scatter(to_scatter, root=root))

        # Wait
        self.comm.barrier()

        # deserialize
        scattered_fixed = scattered[0]
        for scatter in scattered[1:]:
            scattered_fixed += scatter
        new_data = pickle.loads(scattered_fixed)

        # Wait
        self.comm.barrier()

        return new_data

    def batch_gather(self, data, root=0, limit=2 ** 30):
        """Gathers data to all workers in a batched manner that will not exceed the MPI integer limit"""

        # Create local set of data
        serial = pickle.dumps(data, 0)
        num_divisors = self.comm.scatter([len(serial) * self.size // limit + 1] * self.size, root=root)
        final_data = []
        for j in range(num_divisors):  # One for every worker
            start = j * len(serial) // num_divisors
            end = (
                (j + 1) * len(serial) // num_divisors
                if not j == num_divisors - 1
                else (j + 2) * len(serial) // num_divisors
            )

            # Gather
            self.comm.barrier()
            worker_results = self.comm.gather(serial[start:end], root=root)

            # If root, then combine bytes
            if self.rank == root:
                if not len(final_data):
                    final_data = worker_results
                else:
                    final_data = [final_data[i] + d for i, d in enumerate(worker_results)]

        # Wait
        self.comm.barrier()

        # Deserialize
        if self.rank == root:
            return [pickle.loads(i) for i in final_data]
        else:
            return None

    def _run_parallel_functions(self, *tasks) -> list:
        """Takes a series of "tasks" as arguments and returns a list of "results" after running in parallel.

        Arguments
        ---------
        *tasks : tuple
            Each "task" should take the following form: (function, list of arguments, dictionary of kwarguments) and the function will magically compute them in parallel

        Returns
        -------
        list
            Each "result" in the return list will be whatever is returned by the function in the corresponding task
        """

        # Initialize empty completed task list
        finished_tasks = []
        finished_tasks_collective = []

        # Complete all tasks and tag based on initial order for either parallel or not
        if not self.parallel:
            # Linearly execute tasks
            for i, a in tqdm(enumerate(tasks), disable=self.quiet):
                func, arguments, kwarguments = a
                finished_tasks.append(func(*arguments, **kwarguments))
        else:

            # Create data
            if self.am_master():
                data = []
                for j in range(self.size):
                    subdata = []
                    for i, a in enumerate(tasks):
                        if self._should_compute(i, j, self.size):
                            subdata.append((i, a))
                    data.append(subdata)
            else:
                data = None

            # Scatter data
            data = self.batch_scatter(data, root=0)
            new_data = []

            # Compute data
            for i, k in tqdm(enumerate(data), disable=self.quiet):
                index, task = k
                func, arguments, kwarguments = task
                new_data.append((index, func(*arguments, **kwarguments)))

            # Wait until everyone is finished
            self.comm.Barrier()

            # Gather data
            # finished_tasks_collective = self.comm.allgather(new_data)
            finished_tasks_collective = self.batch_gather(new_data, root=0)
            finished_tasks_collective = finished_tasks_collective if self.am_master() else []
            self.comm.Barrier()
            for row in finished_tasks_collective:
                finished_tasks += row

            # Wait until everyone is finished
            self.comm.Barrier()

            # Sort by the tag to return in the original order
            finished_tasks = [i for i in finished_tasks if len(i)]
            finished_tasks = sorted(finished_tasks, key=lambda x: x[0])
            finished_tasks = [i[1] for i in finished_tasks]

        return finished_tasks

    def _should_compute(self, i: int, rank: int, size: int) -> bool:
        """Returns whether or not the rank (process) should be used to complete the given task

        Parameters
        ----------
        i : int
            index of the task
        rank : int
            process rank
        size : int
            number of ranks

        Returns
        -------
        bool:
            true if the process should compute the task
        """
        return i % size == rank

    def _configure_parallel_resources(self) -> None:
        """Sets up the resources for parallel computing"""

        # Use mpi4py to gather resources
        self.comm = MPI.COMM_WORLD
        self.size = self.comm.Get_size()
        self.rank = self.comm.Get_rank()
        if self.am_master():
            print("Running in parallel on {} cores".format(self.size))

    def _propagate_n_only(self) -> None:
        """Propagates all the layers but only gathers the refractive index profiles"""

        # Forward through the device
        m = self.monitors[0] if len(self.monitors) else self.custom_monitors[0]
        cur_len = 0
        for per in range(self.num_periods):
            for layer in tqdm(self.layers, disable=self.quiet):

                # Get system params
                z_list = m.get_z_list(cur_len, cur_len + layer.length)
                n = layer.mode_solver.n

                # Iterate through z
                for i, z in z_list:
                    self._set_monitor(m, i, {"n": n}, n=True)

                cur_len += layer.length

    def _update_state(self, state: int) -> None:
        """Updates the state machine to the provided state

        Parameters
        ----------
        state : int
            state to move to
        """

        self.state = state
        if self.am_master() and not self.quiet:
            print("current state: {}".format(self.states[self.state]))

    def _build_input_array(self, left_coeffs: list, right_coeffs: list, model: Model, m: Monitor) -> dict:
        """Builds the properly formatted input array to be used to calculate field profiles from s matrices and mode coefficients

        Parameters
        ----------
        left_coeffs : list
            A list of floats that represent the mode coefficients for the left side of the full geometry
        right_coeffs : list
            A list of floats that represent the mode coefficients for the right side of the full geometry
        model : simphony.models.Model
            The simphony model base class for storing pins and scattering parameters that will be used to multiply the resulting input array

        Returns
        -------
        dict
            a mapping of the pin names found in the model to their corrisponding weights (mode coefficients) for the field profiles
        """

        # Case 1: left_coeffs > num_modes
        if len(left_coeffs) > self.activated_layers[0][0].num_modes:
            raise Exception("Too many mode coefficients in the left input")

        # Case 2: right_coeffs > num_modes
        if len(left_coeffs) > self.activated_layers[0][-1].num_modes:
            raise Exception("Too many mode coefficients in the right input")

        # Start mapping
        mapping = {}

        # Get sources
        sources = [x for i in self.get_sources().values() for x in i if x in m.sources]

        # Form mapping
        try:
            for pin in model.pins:
                n = pin.name

                # Left global input
                if "left" in n and "dup" not in n:
                    ind = int(n[4:])
                    if ind < len(left_coeffs):
                        mapping[n] = left_coeffs[ind]
                    else:
                        mapping[n] = 0.0

                # Right global input
                if "right" in n and "dup" not in n:
                    ind = int(n[5:])
                    if ind < len(right_coeffs):
                        mapping[n] = right_coeffs[ind]
                    else:
                        mapping[n] = 0.0

                # Left monitor
                if "left" in n and "dup" in n and "to" not in n:
                    mapping[n] = 0

                # Right monitor
                if "right" in n and "dup" in n and "to" not in n:
                    mapping[n] = 0

                # Custom left source inputs
                if "left" in n and "dup" in n and "to" in n:
                    n_ = n.split("_")
                    l, _ = (n_[2], n_[4])
                    ind = int(n_[1].replace("dup", ""))
                    for i, s in enumerate(sources):
                        if s.match_label(l) and ind < len(s.mode_coeffs):
                            mapping[n] = s.mode_coeffs[ind]
                            break
                        elif i == len(sources) - 1:
                            mapping[n] = 0.0

                # Custom right source inputs
                if "right" in n and "dup" in n and "to" in n:
                    n_ = n.split("_")
                    _, r = (n_[2], n_[4])
                    ind = int(n_[1].replace("dup", ""))
                    for i, s in enumerate(sources):
                        if s.match_label(r) and ind < len(s.mode_coeffs):
                            mapping[n] = s.mode_coeffs[ind]
                            break
                        elif i == len(sources) - 1:
                            mapping[n] = 0.0

        except Exception as e:
            print(e)
            raise Exception("Improper format of sources")

        return mapping

    def _layer_field_propagate_part1(self, i: int, l: Model, per: int, cur_len: float) -> list:
        """Propagates the fields through the current layer only. Implements the "field spider/peaker" technique of extracting fields at an arbitrary location without approximating the reflections and actually finding the fully cascaded values.

        Parameters
        ----------
        i : int
            The index of the current layer in the period
        l : Model
            The model that represents the solved layer
        per : int
            The index of the period
        left_coeffs : list
            A list of floats that represent the mode coefficients for the left side of the full geometry
        right_coeffs : list
            A list of floats that represent the mode coefficients for the right side of the full geometry
        cur_len : float
            The current length within the geometry (As it iterates through all the layers)
        z_list : list
            A list of the z points that matter for this layer

        Returns
        -------
        list
            A list of the fields for each point in the input z_list

        """

        # Create output
        activated_layers = (
            self.activated_layers[per] if self.activated_layers[per] is not None else self.activated_layers[0]
        )

        # Period length
        perf = periodic_duplicate_format
        per_length = self._get_total_length()

        # get SP0
        SP0 = []
        for p in range(per):
            network = self.networks[p] if self.networks[p] is not None else self.networks[0]
            interface = (
                self.periodic_interfaces[p] if self.periodic_interfaces[p] is not None else self.periodic_interfaces[0]
            )
            SP0.append(perf(network, p * per_length, (p + 1) * per_length))
            SP0.append(interface)

        # get SP1
        SP1 = []
        for p in range(self.num_periods - (per) - 1):
            p = per + p + 1
            network = self.networks[p] if self.networks[p] is not None else self.networks[0]
            interface = (
                self.periodic_interfaces[p] if self.periodic_interfaces[p] is not None else self.periodic_interfaces[0]
            )
            SP1.append(interface)
            SP1.append(perf(network, p * per_length, (p + 1) * per_length))

        # Get S0
        S0, S0_length = ([], per * per_length)
        for lay in activated_layers[:i]:
            S0.append(perf(lay, S0_length, S0_length + lay.length))
            S0_length += lay.length

        # Get S1
        S1, S1_length = ([], S0_length + activated_layers[i].length)
        for lay in activated_layers[i + 1 :]:
            S1.append(perf(lay, S1_length, S1_length + lay.length))
            S1_length += lay.length

        # See if need to remove sources for periodic l
        checked_l = perf(l, cur_len, cur_len + l.length)

        # Distance params
        num_non_dups = (
            len([i for i in checked_l.pins if "right" in i.name and "dup" not in i.name])
            if (sum(["_to_" in pin.name and "left" in pin.name for pin in checked_l.pins]))
            else len([i for i in checked_l.pins if "left" in i.name and "dup" not in i.name])
        )
        dup = Duplicator(l.wavelength, num_non_dups)

        # create all prop layers
        prop = (
            [*SP0, *S0, checked_l, dup, *S1, *SP1]
            if sum(["_to_" in pin.name and "left" in pin.name for pin in checked_l.pins])
            else [*SP0, *S0, dup, checked_l, *S1, *SP1]
        )

        return prop, checked_l

    def _layer_field_propagate_part2(
        self,
        i: int,
        l: Model,
        left_coeffs: list,
        right_coeffs: list,
        cur_len: float,
        z_list: list,
        m: "Monitor",
        S,
        checked_l_pins,
    ) -> list:

        # Get input array
        result_list = []
        input_map = self._build_input_array(left_coeffs, right_coeffs, S, m)
        coeffs_ = compute(S, input_map, 0)
        coeff_left = np.zeros(len(l.modes), dtype=complex)
        coeff_right = np.zeros(len(l.modes), dtype=complex)
        modes = np.array([[i.Ex, i.Ey, i.Ez, i.Hx, i.Hy, i.Hz] for i in l.modes])
        for i in range(len(l.modes)):
            coeff_left[i] = 0
            coeff_right[i] = 0
            if "left_dup{}".format(i) in coeffs_:
                coeff_left[i] += coeffs_["left_dup{}".format(i)]
            if "right_dup{}".format(i) in coeffs_:
                coeff_right[i] += coeffs_["right_dup{}".format(i)]

        # Case where layer has size 0 (when source at edge of a layer)
        if not len(z_list):
            return result_list

        # Reverse phase if looking from right end
        diffs = [z_list[0] - cur_len] + np.diff(z_list).tolist()
        eig = (2 * np.pi) * np.array([mode.neff for mode in l.modes]) / (self.wavelength)
        if sum(["_to_" in pin.name and "left" in pin.name for pin in checked_l_pins]):
            coeff_left *= np.exp(1j * eig * l.length)
            coeff_right *= np.exp(-1j * eig * l.length)

        # Iterate through z
        for z_diff in diffs:

            # Get coe
            phase_left = np.exp(-z_diff * 1j * eig)
            phase_right = np.exp(z_diff * 1j * eig)
            coeff_left = coeff_left * phase_left
            coeff_right = coeff_right * phase_right
            coeff = coeff_left + coeff_right

            # Create field
            fields_ = modes * coeff[:, np.newaxis, np.newaxis, np.newaxis]
            results = {}
            results["Ex"], results["Ey"], results["Ez"], results["Hx"], results["Hy"], results["Hz"] = fields_.sum(0)
            results["n"] = l.modes[0].n
            result_list.append(results)

        return result_list

    def _set_monitor(self, m: Monitor, i: int, results: dict, n: bool = False) -> None:
        """Adds provided field data to the provides monitor

        Parameters
        ----------
        m : Monitor
            the monitor which to add field data to
        i : int
            the index within the monitor which to add field data
        results : dict
            a dictionary that maps field types (i.e. "Ex", "Hy", "n", etc.) to the field values as a numpy array
        n : bool
            if true, will only care about adding refractive index data to the monitor. Used for propagate_n_only and draw
        """

        # Iterate through all the results
        for key, field in results.items():

            # Really we care about where in the monitor array the field represents
            key = ["Ex", "Ey", "Ez", "Hx", "Hy", "Hz", "n"].index(key) if not n else 0

            # Non implemented fields
            if m.axes in ["x"]:
                raise Exception("x not implemented")
            elif m.axes in ["y"]:
                raise Exception("y not implemented")
            elif m.axes in ["z"]:
                raise Exception("z not implemented")

            # Implemented fields
            if m.axes in ["xy", "yx"]:
                m[key, :, :] = field[:, :]
            elif m.axes in ["xz", "zx"]:
                m[key, :, i] = field[:, int(len(field) / 2)] if field.ndim > 1 else field[:]
            elif m.axes in ["yz", "zy"]:
                m[key, :, i] = field[int(len(field) / 2), :] if field.ndim > 1 else field[:]
            elif m.axes in ["xyz", "xzy", "yxz", "yzx", "zxy", "zyx"]:
                m[key, :, :, i] = field[:, :]

    def _assign_monitor_sources(self, left_coeffs: list = [], right_coeffs: list = []):
        for monitor in self.monitors + self.custom_monitors:
            monitor.left_source = True if len(left_coeffs) else False
            monitor.right_source = True if len(right_coeffs) else False

    def _get_total_length(self) -> float:
        """Returns the total length of a single period of the device"""
        return np.sum([layer.length for layer in self.layers])



import numpy as np
import numpy
from matplotlib import pyplot as plt
import matplotlib
from scipy.interpolate import griddata
plt.rcParams.update({'figure.max_open_warning': 0})


class Monitor(object):
    """Monitor objects store fields during propagation for user visualization. Three types of monitors exist: 3D, 2D, and 1D."""

    def __init__(
        self,
        axes: str = "xz",
        dimensions: tuple = (1, 1),
        components: list = ["E"],
        z_range: tuple = None,
        grid_x: np.array = None,
        grid_y: np.array = None,
        grid_z: np.array = None,
        location: float = None,
        sources: list = [],
        adjoint_n: bool = True,
        total_length: float = 0.0,
    ) -> None:
        """Monitor class constructor0

        Parameters
        ----------
        axes : string
            the spacial axes to capture fields in. Options : 'xz' (default), 'xy', 'yz', 'xyz', 'x', 'y', 'z'. Note, propagation is always in z. (default: "xy")
        dimensions : tuple
            the spacial dimensions of the resulting field (default: (1,1))
        components : list
            list of the field components to store from ('E','H','Ex','Ey','Ez','Hx','Hy','Hz) (default: ["E"])
        z_range : tuple
            tuple or list of the form (start, end) representing the range of the z values to extract (default: None)
        grid_x : numpy array (default: None)
            1d x grid
        grid_y : numpy array (default: None)
            1d y grid
        grid_z : numpy array (default: None)
            1d z grid
        location : float
            the location in z if the monitor represents "xy" axes (default: None)
        sources : list[Source]
            sources to use for the monitor (default:[])
        adjoint_n : bool
            if true will use the "continuous" n used for adjoint
        """

        # Ensure z range is in proper format
        if not (axes in ["xy", "yx"]):
            try:
                if z_range is None:
                    self.start, self.end = [grid_z[0], grid_z[-1]]
                else:
                    self.start, self.end = z_range
            except Exception as e:
                raise Exception(
                    "z_range should be a tuple or list of the form (start, end)"
                    " representing the range of the z values to extract where start "
                    "and end are floats such as (0, 1) for a 1 µm range"
                ) from e

        # Check axes
        if axes == "xz" or axes == "zx":
            self.axes = "xz"
        elif axes == "yz" or axes == "zy":
            self.axes = "yz"
        elif axes in ["xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
            self.axes = "xyz"
        elif axes in ["xy", "yx"]:
            self.axes = "xy"
        else:
            raise Exception(
                f"Monitor setup {axes} has not yet been implemented. Please choose from"
                " the following implemented monitor types: ['xy','yz','xz','xyz']"
            )

        # Set parameters globally
        self.adjoint_n = adjoint_n
        self.dimensions = dimensions
        self.field = np.zeros(dimensions).astype(complex)
        self.lengths = grid_z.tolist()
        self.sources = sources
        self.left_source = False
        self.right_source = False
        self.components = components
        self.layers = {}
        self.grid_x = grid_x
        self.grid_y = grid_y
        self.grid_z = grid_z
        self.location = location
        self.total_length = total_length
        self.xy_monitors = []

    def reset_monitor(self) -> None:
        """Resets the fields in the monitor"""
        self.field *= 0
        # self.sources = []
        # self.left_source = False
        # self.right_source = False

    def get_z_list(self, start: float, end: float) -> list:
        """Finds all the points in z between start and end

        Parameters
        ----------
        start : float
            starting point in z
        end : float
            ending point in z

        Returns
        -------
        list[tuples]
            A list of tuples that take the format (i, l) where i is the index of the z point and l is the z point for all z points in the range
        """
        return [(i, l) for i, l in enumerate(self.lengths) if start <= l <= end]

    def normalize(self) -> None:
        """Normalizes the entire field to 1"""
        self.field[:-1] /= 1  # np.max(np.abs(self.field[:-1, :, 0]))

    def get_array(
        self,
        component: str = "Hy",
        axes: str = None,
        location: float = None,
        z_range: tuple = None,
        grid_x: np.array = None,
        grid_y: np.array = None,
    ) -> "numpy.ndarray":
        """Creates a matplotlib axis displaying the provides field component

        Parameters
        ----------
        component : str
            field component from "['Ex','Ey','Ez','Hx','Hy','Hz','E','H']"
        axes : str
            the spacial axes to capture fields in. Options : 'xz' (default), 'xy', 'yz', 'xyz', 'x', 'y', 'z'. Note, propagation is always in z.
        location : float
            if taken from 3D fields, users can specify where to take their 2D slice. If axes is 'xz', location refers to the location in y and 'yz' refers to a location in x and 'xy' refers to a location in z
        z_range : tuple
            tuple or list of the form (start, end) representing the range of the z values to extract
        grid_x : numpy array
            custom x grid to interpolate onto
        grid_y : numpy array
            custom y grid to interpolate onto

        Returns
        -------
        numpy array
            the requested field
        """

        # Default axes is created upon class creation
        if axes is None:
            axes = self.axes

        # Cannot get plane that was never stored during simulation
        if (self.axes != axes) and (self.axes != "xyz"):
            raise Exception(
                "Different 2D plane was stored during simulation than is being requested. Stored axes: {} different from Requested axes: {}".format(
                    self.axes, axes
                )
            )

        if axes in ["xz", "zx", "yz", "zy", "xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
            # Ensure z range is in proper format
            try:
                if z_range is None:
                    start, end = [self.start, self.end]
                else:
                    start, end = z_range
            except Exception as e:
                raise Exception(
                    "z_range should be a tuple or list of the form (start, end)"
                    "representing the range of the z values to extract where start"
                    "and end are floats such as (0, 1) for a 1 µm range"
                ) from e

            # Get start and end
            def difference_start(list_value):
                return abs(list_value - start)

            def difference_end(list_value):
                return abs(list_value - end)

            s = self.lengths.index(min(self.lengths, key=difference_start))
            e = self.lengths.index(min(self.lengths, key=difference_end)) + 1
            default_grid_z = self.lengths[s:e]

            def m(list_value):
                return abs(list_value - self.grid_z[0])

            m = self.lengths.index(min(self.lengths, key=m))
            s -= m
            e -= m

        elif axes in ["xy", "yx"]:
            default_grid_z = self.grid_z
        else:
            raise Exception("Incorrect axes format")

        # Get default x, y, z grids
        default_grid_x = self.grid_x
        default_grid_y = self.grid_y

        interp_x = False if (grid_x is None) else True
        interp_y = False if (grid_y is None) else True

        # Identify components; components=['Ex','Ey','Ez','Hx','Hy','Hz','n','E','H']
        results = {}
        for i, c in zip(range(len(self.components)), self.components):

            # Perform if needed interpolations and place components
            if self.axes == "xyz":
                # xz plane
                if axes in ["xz", "zx"]:
                    if location:

                        def d(list_value):
                            return abs(list_value - location)

                        index = list(default_grid_x).index(min(default_grid_x, key=d))
                    else:
                        index = int(len(self.field[i][0]) / 2)
                    results[c] = self.field[i][:, index, s:e]

                # yz plane
                elif axes in ["yz", "zy"]:
                    if location:

                        def d(list_value):
                            return abs(list_value - location)

                        index = list(default_grid_y).index(min(default_grid_y, key=d))
                    else:
                        index = int(len(self.field[i]) / 2)
                    results[c] = self.field[i][index, :, s:e]

                # xy plane
                elif axes in ["xy", "yx"]:
                    if location:

                        def d(list_value):
                            return abs(list_value - location)

                        index = list(default_grid_z).index(min(default_grid_z, key=d))
                    else:
                        index = 0
                    results[c] = self.field[i][:, :, index]

                # xyz volume
                elif axes in ["xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
                    results[c] = self.field[i][:, :, s:e]

            elif self.axes in ["xz", "yz", "zx", "zy"]:
                results[c] = self.field[i][:, s:e]

            elif self.axes in ["xy", "yx"]:
                results[c] = self.field[i][:, :]

        # Create E and H fields
        if component == "E":
            results["E"] = np.abs(results["Ex"]) ** 2 + np.abs(results["Ey"]) ** 2 + np.abs(results["Ez"]) ** 2
        if component == "H":
            results["H"] = np.abs(results["Hx"]) ** 2 + np.abs(results["Hy"]) ** 2 + np.abs(results["Hz"]) ** 2

        # List to return
        grid_field = []

        # Custom 2D interpolation function
        def custom_interp2d(field, old_a, old_b, new_a, new_b):
            aa, bb = np.meshgrid(new_a, new_b)
            aa_old, bb_old = np.meshgrid(old_a, old_b)
            points = np.array((aa_old.flatten(), bb_old.flatten())).T
            real = griddata(points, np.real(field).flatten(), (aa, bb)).astype(np.complex128)
            imag = griddata(points, np.real(field).flatten(), (aa, bb)).astype(np.complex128)
            return real + 1j * imag

        # Custom 3D interpolation function
        def custom_interp3d(field, old_a, old_b, old_c, new_a, new_b, new_c):
            aa, bb, cc = np.meshgrid(new_a, new_b, new_c)
            aa_old, bb_old, cc_old = np.meshgrid(old_a, old_b, old_c)
            points = np.array((aa_old.flatten(), bb_old.flatten(), cc_old.flatten())).T
            return griddata(points, np.real(field), (aa, bb, cc)).astype(np.complex128) + 1j * griddata(
                points, np.real(field), (aa, bb)
            ).astype(np.complex128)

        # Add to return list the grid
        if axes in ["xz", "zx"]:
            x = default_grid_x if not interp_x else grid_x
            z = default_grid_z
            grid_field.append(np.array(x))
            grid_field.append(np.array(z))
            if interp_x:
                results[component] = custom_interp2d(results[component], default_grid_z, default_grid_x, z, x)
        elif axes in ["yz", "zy"]:
            y = default_grid_y if not interp_y else grid_y
            z = default_grid_z
            grid_field.append(np.array(y))
            grid_field.append(np.array(z))
            if interp_y:
                results[component] = custom_interp2d(results[component], default_grid_z, default_grid_y, z, y)
        elif axes in ["xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
            x = default_grid_x if not interp_x else grid_x
            y = default_grid_y if not interp_y else grid_y
            z = default_grid_z
            grid_field.append(np.array(x))
            grid_field.append(np.array(y))
            grid_field.append(np.array(z))
            if interp_x or interp_y:
                results[component] = custom_interp3d(
                    results[component], default_grid_x, default_grid_y, default_grid_z, x, y, z
                )
        elif axes in ["xy", "yx"]:
            x = default_grid_x if not interp_x else grid_x
            y = default_grid_y if not interp_y else grid_y
            grid_field.append(np.array(x))
            grid_field.append(np.array(y))
            if interp_x or interp_y:
                results[component] = custom_interp2d(results[component], default_grid_x, default_grid_y, x, y)
        else:
            raise Exception("Please choose valid axes")

        # Add to return list the field
        grid_field.append(np.array(results[component]))

        return grid_field

    def get_source_visual(self, min, max) -> "numpy.ndarray":
        """Returns a mask with lines indicating where a source is"""
        srcs = []
        if self.right_source:
            srcx = [self.total_length, self.total_length]
            srcy = [min, max]
            srcs.append((srcx, srcy))
        if self.left_source:
            srcx = [0, 0]
            srcy = [min, max]
            srcs.append((srcx, srcy))
        for source in self.sources:
            srcx = [source.z, source.z]
            srcy = [min, max]
            srcs.append((srcx, srcy))

        return srcs

    def get_xy_monitor_visual(self, min, max) -> "numpy.ndarray":
        """Returns a mask with lines indicating where a source is"""
        xy_monitors = []
        for monitor in self.xy_monitors:
            srcx = [monitor.location, monitor.location]
            srcy = [min, max]
            xy_monitors.append((srcx, srcy))

        return xy_monitors

    def visualize(
        self,
        ax: matplotlib.image.AxesImage = None,
        component: str = "Hy",
        axes: str = None,
        location: float = 0,
        z_range: tuple = None,
        show_geometry: bool = True,
        show_sources: bool = True,
        show_xy_monitors: bool = False,
    ) -> "matplotlib.image.AxesImage":
        """Creates a matplotlib axis displaying the provides field component

        Parameters
        ----------
        ax : matplotlib axis
            the axis object created when calling plt.figure() or plt.subplots(), if None (default) then the plt interface will be used
        component : string
            field component from "['Ex','Ey','Ez','Hx','Hy','Hz','E','H']"
        axes : string
            the spacial axes to capture fields in. Options : 'xz' (default), 'xy', 'yz', 'xyz', 'x', 'y', 'z'. Note, propagation is always in z.
        location : float
            if taken from 3D fields, users can specify where to take their 2D slice. If axes is 'xz', location refers to the location in y and 'yz' refers to a location in x and 'xy' refers to a location in z.
        z_range : tuple
            tuple or list of the form (start, end) representing the range of the z values to extract
        show_geometry : bool
            if true, will display the geometry faintly under the field profiles (default: True)
        show_sources : bool
            if true, will display a red line indicating where a source is (default: True)

        Returns
        -------
        matplotlib.image.AxesImage
            the image used to plot the index profile
        """

        # Only 2D fields can be generated
        if axes in ["xyz", "yxz", "xzy", "yzx", "zxy", "zyx"]:
            raise Exception(
                "3D fields can be extracted using get_array or visualized on a 2D plane via an axes of xz or yz"
            )

        if axes is None:
            axes = self.axes

        if axes in ["xz", "zx"]:
            axes = "xz"
        elif axes in ["yz", "zy"]:
            axes = "yz"
        elif axes in ["xy", "yx"]:
            axes = "xy"
        else:
            raise Exception("Incorrect axes format")

        yn, zn, n = self.get_array(component="n", axes=axes, location=location, z_range=z_range)
        y, z, field = self.get_array(component=component, axes=axes, location=location, z_range=z_range)

        # Color map lookup table
        cmap_lookup = {
            "Ex": "RdBu",
            "Ey": "RdBu",
            "Ez": "RdBu",
            "Hx": "RdBu",
            "Hy": "RdBu",
            "Hz": "RdBu",
            "E": "Blues",
            "H": "Blues",
            "n": "Greys",
        }

        # Create plots
        if axes in ["xz", "zx", "yz", "zy"]:

            # Underlay the geometry
            show = ax if ax else plt
            if show_geometry:
                show.imshow(
                    np.real(n[::-1]),
                    extent=[np.real(zn[0]), np.real(zn[-1]), np.real(yn[0]), np.real(yn[-1])],
                    cmap=cmap_lookup["n"],
                )
            vmin, vmax = (np.real(np.min(field)), np.real(np.max(field)))
            alpha = 1 if not show_geometry else 0.85

            # Plot sources
            if show_sources:
                srcs = self.get_source_visual(np.real(y[0]), np.real(y[-1]))
                for srcx, srcy in srcs:
                    plt.plot(srcx, srcy, color="red", linewidth=2)

            # Plot xy monitors
            if show_xy_monitors:
                xy_monitors = self.get_xy_monitor_visual(np.real(y[0]), np.real(y[-1]))
                for locx, locy in xy_monitors:
                    plt.plot(locx, locy, color="blue", linewidth=2)

            im = show.imshow(
                np.real(field[::-1]),
                extent=[np.real(z[0]), np.real(z[-1]), np.real(y[0]), np.real(y[-1])],
                cmap=cmap_lookup[component],
                alpha=alpha,
                vmin=vmin,
                vmax=vmax,
            )

            # Assign labels
            if not ax:
                plt.xlabel(np.real(axes[1]))
                plt.ylabel(np.real(axes[0]))
                plt.title(component)
            else:
                ax.set_xlabel(np.real(axes[1]))
                ax.set_ylabel(np.real(axes[0]))
                ax.set_title(component)

        # xy fields
        else:
            if ax:
                im = ax.imshow(
                    np.rot90(np.real(field)),
                    extent=[np.real(z[0]), np.real(z[-1]), np.real(y[0]), np.real(y[-1])],
                    cmap=cmap_lookup[component],
                )
                ax.set_xlabel(np.real(axes[0]))
                ax.set_ylabel(np.real(axes[1]))
                ax.set_title(component)
            else:
                im = plt.imshow(
                    np.rot90(np.real(field)),
                    extent=[np.real(z[0]), np.real(z[-1]), np.real(y[0]), np.real(y[-1])],
                    cmap=cmap_lookup[component],
                )
                plt.xlabel(np.real(axes[0]))
                plt.ylabel(np.real(axes[1]))
                plt.title(component)

        return im

    def __getitem__(self, subscript):
        return self.field[subscript]

    def __setitem__(self, subscript, item):
        self.field[subscript] = item

    def __delitem__(self, subscript):
        del self.field[subscript]



"""
emepy

Eigenmode Expansion Python
"""

__version__ = "1.2.1"
__author__ = "Ian Hammond"
__credits__ = "BYU CamachoLab"

from emepy.eme import *
from emepy.tools import *
from emepy.mode import *
from emepy.fd import *
from emepy.models import *
from emepy.source import *
from emepy.ann import *
from emepy.monitors import *
from emepy.lumerical import *
from emepy.materials import *
